@@ -7,10 +7,10 @@ use core::ffi::{c_char, c_void, CStr};
 use core::{mem, ptr};
 
 use nginx_sys::{
-    ngx_command_t, ngx_conf_parse, ngx_conf_t, ngx_http_core_srv_conf_t, ngx_str_t, ngx_uint_t,
-    NGX_CONF_1MORE, NGX_CONF_BLOCK, NGX_CONF_FLAG, NGX_CONF_NOARGS, NGX_CONF_TAKE1,
-    NGX_HTTP_MAIN_CONF, NGX_HTTP_MAIN_CONF_OFFSET, NGX_HTTP_SRV_CONF, NGX_HTTP_SRV_CONF_OFFSET,
-    NGX_LOG_EMERG,
+    ngx_command_t, ngx_conf_parse, ngx_conf_t, ngx_decode_base64url, ngx_http_core_srv_conf_t,
+    ngx_str_t, ngx_uint_t, NGX_CONF_1MORE, NGX_CONF_BLOCK, NGX_CONF_FLAG, NGX_CONF_NOARGS,
+    NGX_CONF_TAKE1, NGX_CONF_TAKE2, NGX_HTTP_MAIN_CONF, NGX_HTTP_MAIN_CONF_OFFSET,
+    NGX_HTTP_SRV_CONF, NGX_HTTP_SRV_CONF_OFFSET, NGX_LOG_EMERG,
 };
 use ngx::collections::Vec;
 use ngx::core::{Pool, Status, NGX_CONF_ERROR, NGX_CONF_OK};
@@ -22,6 +22,8 @@ use self::issuer::Issuer;
 use self::order::CertificateOrder;
 use self::pkey::PrivateKey;
 use self::shared_zone::{SharedZone, ACME_ZONE_NAME, ACME_ZONE_SIZE};
+use self::ssl::NgxSsl;
+use crate::acme::types::ChallengeKind;
 use crate::state::AcmeSharedData;
 
 pub mod ext;
@@ -33,11 +35,14 @@ pub mod shared_zone;
 pub mod ssl;
 
 const NGX_CONF_DUPLICATE: *mut c_char = c"is duplicate".as_ptr().cast_mut();
+const NGX_CONF_INVALID_VALUE: *mut c_char = c"invalid value".as_ptr().cast_mut();
+pub const NGX_CONF_UNSET_PTR: *mut core::ffi::c_void = nginx_sys::NGX_CONF_UNSET as _;
 
 /// Main (http block) level configuration.
 #[derive(Debug, Default)]
 pub struct AcmeMainConfig {
     pub issuers: Vec<Issuer>,
+    pub ssl: NgxSsl,
     pub data: Option<&'static AcmeSharedData>,
     pub shm_zone: shared_zone::SharedZone,
 }
@@ -48,7 +53,7 @@ pub struct AcmeServerConfig {
     pub issuer: ngx_str_t,
     // Only one certificate order per server block is currently allowed. For multiple entries we
     // will have to implement certificate selection in the variable handler.
-    pub order: Option<CertificateOrder<ngx_str_t, Pool>>,
+    pub order: Option<CertificateOrder<&'static str, Pool>>,
 }
 
 pub static mut NGX_HTTP_ACME_COMMANDS: [ngx_command_t; 4] = [
@@ -79,7 +84,7 @@ pub static mut NGX_HTTP_ACME_COMMANDS: [ngx_command_t; 4] = [
     ngx_command_t::empty(),
 ];
 
-static mut NGX_HTTP_ACME_ISSUER_COMMANDS: [ngx_command_t; 8] = [
+static mut NGX_HTTP_ACME_ISSUER_COMMANDS: [ngx_command_t; 12] = [
     ngx_command_t {
         name: ngx_string!("uri"),
         type_: NGX_CONF_TAKE1 as ngx_uint_t,
@@ -96,6 +101,14 @@ static mut NGX_HTTP_ACME_ISSUER_COMMANDS: [ngx_command_t; 8] = [
         offset: 0,
         post: ptr::null_mut(),
     },
+    ngx_command_t {
+        name: ngx_string!("challenge"),
+        type_: NGX_CONF_TAKE1 as ngx_uint_t,
+        set: Some(cmd_issuer_set_challenge),
+        conf: 0,
+        offset: 0,
+        post: ptr::null_mut(),
+    },
     ngx_command_t {
         name: ngx_string!("contact"),
         type_: NGX_CONF_TAKE1 as ngx_uint_t,
@@ -104,6 +117,30 @@ static mut NGX_HTTP_ACME_ISSUER_COMMANDS: [ngx_command_t; 8] = [
         offset: 0,
         post: ptr::null_mut(),
     },
+    ngx_command_t {
+        name: ngx_string!("external_account_key"),
+        type_: NGX_CONF_TAKE2 as ngx_uint_t,
+        set: Some(cmd_issuer_set_external_account_key),
+        conf: 0,
+        offset: 0,
+        post: ptr::null_mut(),
+    },
+    ngx_command_t {
+        name: ngx_string!("preferred_chain"),
+        type_: NGX_CONF_TAKE1 as ngx_uint_t,
+        set: Some(cmd_issuer_set_preferred_chain),
+        conf: 0,
+        offset: 0,
+        post: ptr::null_mut(),
+    },
+    ngx_command_t {
+        name: ngx_string!("profile"),
+        type_: nginx_sys::NGX_CONF_TAKE12 as ngx_uint_t,
+        set: Some(cmd_issuer_set_profile),
+        conf: 0,
+        offset: 0,
+        post: ptr::null_mut(),
+    },
     ngx_command_t {
         name: ngx_string!("ssl_trusted_certificate"),
         type_: NGX_CONF_TAKE1 as ngx_uint_t,
@@ -123,7 +160,7 @@ static mut NGX_HTTP_ACME_ISSUER_COMMANDS: [ngx_command_t; 8] = [
     ngx_command_t {
         name: ngx_string!("state_path"),
         type_: NGX_CONF_TAKE1 as ngx_uint_t,
-        set: Some(nginx_sys::ngx_conf_set_path_slot),
+        set: Some(cmd_issuer_set_state_path),
         conf: 0,
         offset: mem::offset_of!(Issuer, state_path),
         post: ptr::null_mut(),
@@ -279,7 +316,7 @@ extern "C" fn cmd_add_certificate(
         return c"\"issuer\" is missing".as_ptr().cast_mut();
     }
 
-    let mut order = CertificateOrder::<ngx_str_t, Pool>::new_in(cf.pool());
+    let mut order = CertificateOrder::<&'static str, Pool>::new_in(cf.pool());
 
     for value in &args[2..] {
         if let Some(key) = value.strip_prefix(b"key=") {
@@ -291,7 +328,17 @@ extern "C" fn cmd_add_certificate(
             continue;
         }
 
-        if let Err(err) = order.try_add_identifier(value) {
+        if value.is_empty() {
+            return NGX_CONF_INVALID_VALUE;
+        }
+
+        // SAFETY: the value is not empty, well aligned, and the conversion result is assigned to an
+        // object in the same pool.
+        let Ok(value) = (unsafe { conf_value_to_str(value) }) else {
+            return NGX_CONF_INVALID_VALUE;
+        };
+
+        if let Err(err) = order.try_add_identifier(cf, value) {
             return cf.error(args[0], &err);
         }
     }
@@ -304,6 +351,35 @@ extern "C" fn cmd_add_certificate(
     NGX_CONF_OK
 }
 
+extern "C" fn cmd_issuer_set_challenge(
+    cf: *mut ngx_conf_t,
+    _cmd: *mut ngx_command_t,
+    conf: *mut c_void,
+) -> *mut c_char {
+    let cf = unsafe { cf.as_mut().expect("cf") };
+    let issuer = unsafe { conf.cast::<Issuer>().as_mut().expect("issuer conf") };
+
+    if issuer.challenge.is_some() {
+        return NGX_CONF_DUPLICATE;
+    }
+
+    // NGX_CONF_TAKE1 ensures that args contains 2 elements
+    let val = cf.args()[1];
+
+    let val = match val.as_bytes() {
+        b"http" | b"http-01" => ChallengeKind::Http01,
+        b"tls-alpn" | b"tls-alpn-01" => ChallengeKind::TlsAlpn01,
+        _ => {
+            ngx_conf_log_error!(NGX_LOG_EMERG, cf, "unsupported challenge: {val}");
+            return NGX_CONF_ERROR;
+        }
+    };
+
+    issuer.challenge = Some(val);
+
+    NGX_CONF_OK
+}
+
 extern "C" fn cmd_issuer_add_contact(
     cf: *mut ngx_conf_t,
     _cmd: *mut ngx_command_t,
@@ -337,12 +413,12 @@ extern "C" fn cmd_issuer_add_contact(
     // NGX_CONF_TAKE1 ensures that args contains 2 elements
     let args = cf.args();
 
-    if args[1].is_empty() || core::str::from_utf8(args[1].as_bytes()).is_err() {
-        return c"invalid value".as_ptr().cast_mut();
+    if args[1].is_empty() {
+        return NGX_CONF_INVALID_VALUE;
     };
 
-    if has_scheme(args[1].as_ref()) {
-        issuer.contacts.push(args[1]);
+    let value = if has_scheme(args[1].as_ref()) {
+        args[1]
     } else {
         let mut value = ngx_str_t::empty();
         value.len = MAILTO.len() + args[1].len;
@@ -353,11 +429,18 @@ extern "C" fn cmd_issuer_add_contact(
 
         value.as_bytes_mut()[..MAILTO.len()].copy_from_slice(MAILTO);
         value.as_bytes_mut()[MAILTO.len()..].copy_from_slice(args[1].as_ref());
+        value
+    };
 
-        issuer.contacts.push(value);
+    // SAFETY: the value is not empty, well aligned, and the conversion result is assigned to an
+    // object in the same pool.
+    match unsafe { conf_value_to_str(&value) } {
+        Ok(x) => {
+            issuer.contacts.push(x);
+            NGX_CONF_OK
+        }
+        Err(_) => NGX_CONF_INVALID_VALUE,
     }
-
-    NGX_CONF_OK
 }
 
 extern "C" fn cmd_issuer_set_account_key(
@@ -383,6 +466,124 @@ extern "C" fn cmd_issuer_set_account_key(
     NGX_CONF_OK
 }
 
+extern "C" fn cmd_issuer_set_external_account_key(
+    cf: *mut ngx_conf_t,
+    _cmd: *mut ngx_command_t,
+    conf: *mut c_void,
+) -> *mut c_char {
+    let cf = unsafe { cf.as_mut().expect("cf") };
+    let issuer = unsafe { conf.cast::<Issuer>().as_mut().expect("issuer conf") };
+
+    if issuer.eab_key.is_some() {
+        return NGX_CONF_DUPLICATE;
+    }
+
+    let pool = cf.pool();
+    // NGX_CONF_TAKE2 ensures that args contains 3 elements
+    let args = cf.args();
+
+    if args[1].is_empty() || args[2].is_empty() {
+        return NGX_CONF_INVALID_VALUE;
+    }
+
+    // SAFETY: the value is not empty, well aligned, and the conversion result is assigned to an
+    // object in the same pool.
+    let Ok(kid) = (unsafe { conf_value_to_str(&args[1]) }) else {
+        return NGX_CONF_INVALID_VALUE;
+    };
+
+    let mut encoded = if let Some(arg) = args[2].strip_prefix(b"data:") {
+        arg
+    } else {
+        match crate::util::read_to_ngx_str(cf, &args[2]) {
+            Ok(x) => x,
+            Err(e) => return cf.error(args[0], &e),
+        }
+    };
+
+    crate::util::ngx_str_trim(&mut encoded);
+
+    let len = encoded.len.div_ceil(4) * 3;
+    let mut key = ngx_str_t {
+        data: pool.alloc_unaligned(len).cast(),
+        len,
+    };
+
+    if key.data.is_null() {
+        return NGX_CONF_ERROR;
+    }
+
+    if !Status(unsafe { ngx_decode_base64url(&mut key, &mut encoded) }).is_ok() {
+        return c"invalid base64url encoded value".as_ptr().cast_mut();
+    }
+
+    issuer.eab_key = Some(issuer::ExternalAccountKey { kid, key });
+
+    NGX_CONF_OK
+}
+
+extern "C" fn cmd_issuer_set_preferred_chain(
+    cf: *mut ngx_conf_t,
+    _cmd: *mut ngx_command_t,
+    conf: *mut c_void,
+) -> *mut c_char {
+    let cf = unsafe { cf.as_mut().expect("cf") };
+    let issuer = unsafe { conf.cast::<Issuer>().as_mut().expect("issuer conf") };
+
+    if issuer.chain.is_some() {
+        return NGX_CONF_DUPLICATE;
+    }
+
+    // NGX_CONF_TAKE1 ensures that args contains 2 elements
+    let args = cf.args();
+
+    // SAFETY: the value is well aligned, and the conversion result is assigned to an object in
+    // the same pool.
+    let Ok(issuer_name) = (unsafe { conf_value_to_str(&args[1]) }) else {
+        return NGX_CONF_INVALID_VALUE;
+    };
+
+    issuer.chain = Some(issuer::CertificateChainMatcher::new(issuer_name));
+
+    NGX_CONF_OK
+}
+
+extern "C" fn cmd_issuer_set_profile(
+    cf: *mut ngx_conf_t,
+    _cmd: *mut ngx_command_t,
+    conf: *mut c_void,
+) -> *mut c_char {
+    let cf = unsafe { cf.as_mut().expect("cf") };
+    let issuer = unsafe { conf.cast::<Issuer>().as_mut().expect("issuer conf") };
+
+    if !matches!(issuer.profile, issuer::Profile::Unset) {
+        return NGX_CONF_DUPLICATE;
+    }
+
+    // NGX_CONF_TAKE12 ensures that args contains either 2 or 3 elements
+    let args = cf.args();
+
+    // SAFETY: the value is not empty, well aligned, and the conversion result is assigned to an
+    // object in the same pool.
+    let Ok(profile) = (unsafe { conf_value_to_str(&args[1]) }) else {
+        return NGX_CONF_INVALID_VALUE;
+    };
+
+    let require = match args.get(2) {
+        Some(x) if x.as_ref() == b"require" => true,
+        Some(_) => return NGX_CONF_INVALID_VALUE,
+        None => false,
+    };
+
+    issuer.profile = if require {
+        issuer::Profile::Required(profile)
+    } else {
+        issuer::Profile::Preferred(profile)
+    };
+
+    NGX_CONF_OK
+}
+
 extern "C" fn cmd_issuer_set_uri(
     cf: *mut ngx_conf_t,
     _cmd: *mut ngx_command_t,
@@ -407,6 +608,47 @@ extern "C" fn cmd_issuer_set_uri(
     NGX_CONF_OK
 }
 
+/// A wrapper over the `ngx_conf_set_path_slot` that takes the "off" value to disable persistency.
+extern "C" fn cmd_issuer_set_state_path(
+    cf: *mut ngx_conf_t,
+    cmd: *mut ngx_command_t,
+    conf: *mut c_void,
+) -> *mut c_char {
+    let cf = unsafe { cf.as_mut().expect("cf ptr is always valid") };
+    let issuer = unsafe { conf.cast::<Issuer>().as_mut().expect("issuer conf") };
+
+    if issuer.state_path != NGX_CONF_UNSET_PTR.cast() {
+        return NGX_CONF_DUPLICATE;
+    }
+
+    issuer.state_path = ptr::null_mut();
+
+    // NGX_CONF_TAKE1 ensures that args contains 2 elements
+    let mut path = cf.args()[1];
+
+    if path.as_bytes() == b"off" {
+        return NGX_CONF_OK;
+    }
+
+    // We need to add our prefix before we pass the path to ngx_conf_set_path_slot,
+    // because otherwise it will be resolved with cycle->prefix.
+    if let Some(p) = issuer::NGX_ACME_STATE_PREFIX {
+        let mut p = ngx_str_t {
+            data: p.as_ptr().cast_mut(),
+            len: p.len(),
+        };
+
+        // ngx_get_full_name does not modify input buffers.
+        if !Status(unsafe { nginx_sys::ngx_get_full_name(cf.pool, &mut p, &mut path) }).is_ok() {
+            return NGX_CONF_ERROR;
+        }
+
+        cf.args_mut()[1] = path;
+    }
+
+    unsafe { nginx_sys::ngx_conf_set_path_slot(cf, cmd, ptr::from_mut(issuer).cast()) }
+}
+
 extern "C" fn cmd_issuer_set_accept_tos(
     _cf: *mut ngx_conf_t,
     _cmd: *mut ngx_command_t,
@@ -566,3 +808,26 @@ fn conf_check_nargs(cmd: &ngx_command_t, nargs: ngx_uint_t) -> bool {
         nargs <= ARGUMENT_NUMBER.len() && (flags & ARGUMENT_NUMBER[nargs - 1]) != 0
     }
 }
+
+/// Unsafely converts `ngx_str_t` into a static UTF-8 string reference.
+///
+/// # Safety
+///
+/// `value` must be allocated on the configuration (cycle) pool, and stored in another object on the
+/// same pool. With that, we can expect that both the borrowed string and the owning object will be
+/// destroyed simultaneously.
+///
+/// In the worker process this happens at the process exit, making the `'static` lifetime specifier
+/// accurate.
+/// In the master process, the cycle pool is destroyed after reloading the configuration, along with
+/// all the configuration objects. But this process role is not capable of serving connections or
+/// running background tasks, and thus will not create additional borrows with potentially extended
+/// lifetime.
+pub unsafe fn conf_value_to_str(value: &ngx_str_t) -> Result<&'static str, core::str::Utf8Error> {
+    if value.len == 0 {
+        Ok("")
+    } else {
+        let bytes = core::slice::from_raw_parts(value.data, value.len);
+        core::str::from_utf8(bytes)
+    }
+}
